@@ -26,12 +26,13 @@ import (
 
 	"github.com/openshift/cluster-api/pkg/apis/machine/v1beta1"
 	fakeclusterapi "github.com/openshift/cluster-api/pkg/client/clientset_generated/clientset/fake"
-	apiv1 "k8s.io/api/core/v1"
-	"k8s.io/apimachinery/pkg/apis/meta/v1"
+	corev1 "k8s.io/api/core/v1"
+	v1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/labels"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/types"
 	fakekube "k8s.io/client-go/kubernetes/fake"
+	"k8s.io/utils/pointer"
 )
 
 type testControllerShutdownFunc func()
@@ -41,7 +42,7 @@ type testConfig struct {
 	machineDeployment *v1beta1.MachineDeployment
 	machineSet        *v1beta1.MachineSet
 	machines          []*v1beta1.Machine
-	nodes             []*apiv1.Node
+	nodes             []*corev1.Node
 }
 
 type testSpec struct {
@@ -50,7 +51,6 @@ type testSpec struct {
 	machineSetName          string
 	namespace               string
 	nodeCount               int
-	replicaCount            int32
 	rootIsMachineDeployment bool
 }
 
@@ -92,23 +92,23 @@ func mustCreateTestController(t *testing.T, testConfigs ...*testConfig) (*machin
 	}
 }
 
-func createMachineSetTestConfig(namespace string, nodeCount int, replicaCount int32, annotations map[string]string) *testConfig {
-	return createTestConfigs(createTestSpecs(namespace, 1, nodeCount, replicaCount, false, annotations)...)[0]
+func createMachineSetTestConfig(namespace string, nodeCount int, annotations map[string]string) *testConfig {
+	return createTestConfigs(createTestSpecs(namespace, 1, nodeCount, false, annotations)...)[0]
 }
 
-func createMachineSetTestConfigs(namespace string, configCount, nodeCount int, replicaCount int32, annotations map[string]string) []*testConfig {
-	return createTestConfigs(createTestSpecs(namespace, configCount, nodeCount, replicaCount, false, annotations)...)
+func createMachineSetTestConfigs(namespace string, configCount, nodeCount int, annotations map[string]string) []*testConfig {
+	return createTestConfigs(createTestSpecs(namespace, configCount, nodeCount, false, annotations)...)
 }
 
-func createMachineDeploymentTestConfig(namespace string, nodeCount int, replicaCount int32, annotations map[string]string) *testConfig {
-	return createTestConfigs(createTestSpecs(namespace, 1, nodeCount, replicaCount, true, annotations)...)[0]
+func createMachineDeploymentTestConfig(namespace string, nodeCount int, annotations map[string]string) *testConfig {
+	return createTestConfigs(createTestSpecs(namespace, 1, nodeCount, true, annotations)...)[0]
 }
 
-func createMachineDeploymentTestConfigs(namespace string, configCount, nodeCount int, replicaCount int32, annotations map[string]string) []*testConfig {
-	return createTestConfigs(createTestSpecs(namespace, configCount, nodeCount, replicaCount, true, annotations)...)
+func createMachineDeploymentTestConfigs(namespace string, configCount, nodeCount int, annotations map[string]string) []*testConfig {
+	return createTestConfigs(createTestSpecs(namespace, configCount, nodeCount, true, annotations)...)
 }
 
-func createTestSpecs(namespace string, scalableResourceCount, nodeCount int, replicaCount int32, isMachineDeployment bool, annotations map[string]string) []testSpec {
+func createTestSpecs(namespace string, scalableResourceCount, nodeCount int, isMachineDeployment bool, annotations map[string]string) []testSpec {
 	var specs []testSpec
 
 	for i := 0; i < scalableResourceCount; i++ {
@@ -118,7 +118,6 @@ func createTestSpecs(namespace string, scalableResourceCount, nodeCount int, rep
 			machineSetName:          fmt.Sprintf("machineset-%d", i),
 			namespace:               strings.ToLower(namespace),
 			nodeCount:               nodeCount,
-			replicaCount:            replicaCount,
 			rootIsMachineDeployment: isMachineDeployment,
 		})
 	}
@@ -132,7 +131,7 @@ func createTestConfigs(specs ...testSpec) []*testConfig {
 	for i, spec := range specs {
 		config := &testConfig{
 			spec:     &specs[i],
-			nodes:    make([]*apiv1.Node, spec.nodeCount),
+			nodes:    make([]*corev1.Node, spec.nodeCount),
 			machines: make([]*v1beta1.Machine, spec.nodeCount),
 		}
 
@@ -149,7 +148,7 @@ func createTestConfigs(specs ...testSpec) []*testConfig {
 
 		if !spec.rootIsMachineDeployment {
 			config.machineSet.ObjectMeta.Annotations = spec.annotations
-			config.machineSet.Spec.Replicas = int32ptr(spec.replicaCount)
+			config.machineSet.Spec.Replicas = int32ptr(int32(spec.nodeCount))
 		} else {
 			config.machineDeployment = &v1beta1.MachineDeployment{
 				TypeMeta: v1.TypeMeta{
@@ -162,7 +161,7 @@ func createTestConfigs(specs ...testSpec) []*testConfig {
 					Annotations: spec.annotations,
 				},
 				Spec: v1beta1.MachineDeploymentSpec{
-					Replicas: int32ptr(spec.replicaCount),
+					Replicas: int32ptr(int32(spec.nodeCount)),
 				},
 			}
 
@@ -193,8 +192,8 @@ func createTestConfigs(specs ...testSpec) []*testConfig {
 // makeLinkedNodeAndMachine creates a node and machine. The machine
 // has its NodeRef set to the new node and the new machine's owner
 // reference is set to owner.
-func makeLinkedNodeAndMachine(i int, namespace string, owner v1.OwnerReference) (*apiv1.Node, *v1beta1.Machine) {
-	node := &apiv1.Node{
+func makeLinkedNodeAndMachine(i int, namespace string, owner v1.OwnerReference) (*corev1.Node, *v1beta1.Machine) {
+	node := &corev1.Node{
 		TypeMeta: v1.TypeMeta{
 			Kind: "Node",
 		},
@@ -204,7 +203,7 @@ func makeLinkedNodeAndMachine(i int, namespace string, owner v1.OwnerReference)
 				machineAnnotationKey: fmt.Sprintf("%s/%s-%s-machine-%d", namespace, namespace, owner.Name, i),
 			},
 		},
-		Spec: apiv1.NodeSpec{
+		Spec: corev1.NodeSpec{
 			ProviderID: fmt.Sprintf("%s-%s-nodeid-%d", namespace, owner.Name, i),
 		},
 	}
@@ -222,8 +221,11 @@ func makeLinkedNodeAndMachine(i int, namespace string, owner v1.OwnerReference)
 				UID:  owner.UID,
 			}},
 		},
+		Spec: v1beta1.MachineSpec{
+			ProviderID: pointer.StringPtr(fmt.Sprintf("%s-%s-nodeid-%d", namespace, owner.Name, i)),
+		},
 		Status: v1beta1.MachineStatus{
-			NodeRef: &apiv1.ObjectReference{
+			NodeRef: &corev1.ObjectReference{
 				Kind: node.Kind,
 				Name: node.Name,
 			},
@@ -336,7 +338,7 @@ func TestControllerFindMachineByID(t *testing.T) {
 
 	for _, tc := range testCases {
 		t.Run(tc.description, func(t *testing.T) {
-			testConfig := createMachineSetTestConfig(testNamespace, 1, 1, map[string]string{
+			testConfig := createMachineSetTestConfig(testNamespace, 1, map[string]string{
 				nodeGroupMinSizeAnnotationKey: "1",
 				nodeGroupMaxSizeAnnotationKey: "10",
 			})
@@ -352,7 +354,7 @@ func TestControllerFindMachineByID(t *testing.T) {
 }
 
 func TestControllerFindMachineOwner(t *testing.T) {
-	testConfig := createMachineSetTestConfig(testNamespace, 1, 1, map[string]string{
+	testConfig := createMachineSetTestConfig(testNamespace, 1, map[string]string{
 		nodeGroupMinSizeAnnotationKey: "1",
 		nodeGroupMaxSizeAnnotationKey: "10",
 	})
@@ -396,8 +398,8 @@ func TestControllerFindMachineOwner(t *testing.T) {
 	}
 }
 
-func TestControllerFindMachineByNodeProviderID(t *testing.T) {
-	testConfig := createMachineSetTestConfig(testNamespace, 1, 1, map[string]string{
+func TestControllerFindMachineByProviderID(t *testing.T) {
+	testConfig := createMachineSetTestConfig(testNamespace, 1, map[string]string{
 		nodeGroupMinSizeAnnotationKey: "1",
 		nodeGroupMaxSizeAnnotationKey: "10",
 	})
@@ -405,9 +407,18 @@ func TestControllerFindMachineByNodeProviderID(t *testing.T) {
 	controller, stop := mustCreateTestController(t, testConfig)
 	defer stop()
 
-	// Test #1: Verify node can be found because it has a
-	// ProviderID value and a machine annotation.
-	machine, err := controller.findMachineByNodeProviderID(testConfig.nodes[0])
+	// Remove all the "machine" annotation values on all the
+	// nodes. We want to force findMachineByProviderID() to only
+	// be successful by searching on provider ID.
+	for _, node := range testConfig.nodes {
+		delete(node.Annotations, machineAnnotationKey)
+		if err := controller.nodeInformer.GetStore().Update(node); err != nil {
+			t.Fatalf("unexpected error updating node, got %v", err)
+		}
+	}
+
+	// Test #1: Verify underlying machine provider ID matches
+	machine, err := controller.findMachineByProviderID(testConfig.nodes[0].Spec.ProviderID)
 	if err != nil {
 		t.Fatalf("unexpected error: %v", err)
 	}
@@ -418,35 +429,24 @@ func TestControllerFindMachineByNodeProviderID(t *testing.T) {
 		t.Fatalf("expected machines to be equal - expected %+v, got %+v", testConfig.machines[0], machine)
 	}
 
-	// Test #2: Verify node is not found if it has a non-existent ProviderID
-	node := testConfig.nodes[0].DeepCopy()
-	node.Spec.ProviderID = ""
-	nonExistentMachine, err := controller.findMachineByNodeProviderID(node)
+	// Test #2: Verify machine is not found if it has a
+	// non-existent or different provider ID.
+	machine = testConfig.machines[0].DeepCopy()
+	machine.Spec.ProviderID = pointer.StringPtr("does-not-match")
+	if err := controller.machineInformer.Informer().GetStore().Update(machine); err != nil {
+		t.Fatalf("unexpected error updating machine, got %v", err)
+	}
+	machine, err = controller.findMachineByProviderID(testConfig.nodes[0].Spec.ProviderID)
 	if err != nil {
 		t.Fatalf("unexpected error: %v", err)
 	}
-	if nonExistentMachine != nil {
-		t.Fatal("expected find to fail")
-	}
-
-	// Test #3: Verify node is not found if the stored object has
-	// no "machine" annotation
-	node = testConfig.nodes[0].DeepCopy()
-	delete(node.Annotations, machineAnnotationKey)
-	if err := controller.nodeInformer.GetStore().Update(node); err != nil {
-		t.Fatalf("unexpected error updating node, got %v", err)
-	}
-	nonExistentMachine, err = controller.findMachineByNodeProviderID(testConfig.nodes[0])
-	if err != nil {
-		t.Fatalf("unexpected error: %v", err)
-	}
-	if nonExistentMachine != nil {
+	if machine != nil {
 		t.Fatal("expected find to fail")
 	}
 }
 
 func TestControllerFindNodeByNodeName(t *testing.T) {
-	testConfig := createMachineSetTestConfig(testNamespace, 1, 1, map[string]string{
+	testConfig := createMachineSetTestConfig(testNamespace, 1, map[string]string{
 		nodeGroupMinSizeAnnotationKey: "1",
 		nodeGroupMaxSizeAnnotationKey: "10",
 	})
@@ -474,7 +474,7 @@ func TestControllerFindNodeByNodeName(t *testing.T) {
 }
 
 func TestControllerMachinesInMachineSet(t *testing.T) {
-	testConfig1 := createMachineSetTestConfig("testConfig1", 5, 5, map[string]string{
+	testConfig1 := createMachineSetTestConfig("testConfig1", 5, map[string]string{
 		nodeGroupMinSizeAnnotationKey: "1",
 		nodeGroupMaxSizeAnnotationKey: "10",
 	})
@@ -486,7 +486,7 @@ func TestControllerMachinesInMachineSet(t *testing.T) {
 	// nodes and the additional machineset to the existing set of
 	// test objects in the controller. This gives us two
 	// machinesets, each with their own machines and linked nodes.
-	testConfig2 := createMachineSetTestConfig("testConfig2", 5, 5, map[string]string{
+	testConfig2 := createMachineSetTestConfig("testConfig2", 5, map[string]string{
 		nodeGroupMinSizeAnnotationKey: "1",
 		nodeGroupMaxSizeAnnotationKey: "10",
 	})
@@ -547,7 +547,7 @@ func TestControllerMachinesInMachineSet(t *testing.T) {
 }
 
 func TestControllerLookupNodeGroupForNonExistentNode(t *testing.T) {
-	testConfig := createMachineSetTestConfig(testNamespace, 1, 1, map[string]string{
+	testConfig := createMachineSetTestConfig(testNamespace, 1, map[string]string{
 		nodeGroupMinSizeAnnotationKey: "1",
 		nodeGroupMaxSizeAnnotationKey: "10",
 	})
@@ -593,7 +593,7 @@ func TestControllerNodeGroupForNodeWithMissingMachineOwner(t *testing.T) {
 	}
 
 	t.Run("MachineSet", func(t *testing.T) {
-		testConfig := createMachineSetTestConfig(testNamespace, 1, 1, map[string]string{
+		testConfig := createMachineSetTestConfig(testNamespace, 1, map[string]string{
 			nodeGroupMinSizeAnnotationKey: "1",
 			nodeGroupMaxSizeAnnotationKey: "10",
 		})
@@ -601,7 +601,7 @@ func TestControllerNodeGroupForNodeWithMissingMachineOwner(t *testing.T) {
 	})
 
 	t.Run("MachineDeployment", func(t *testing.T) {
-		testConfig := createMachineDeploymentTestConfig(testNamespace, 1, 1, map[string]string{
+		testConfig := createMachineDeploymentTestConfig(testNamespace, 1, map[string]string{
 			nodeGroupMinSizeAnnotationKey: "1",
 			nodeGroupMaxSizeAnnotationKey: "10",
 		})
@@ -626,7 +626,7 @@ func TestControllerNodeGroupForNodeWithPositiveScalingBounds(t *testing.T) {
 	}
 
 	t.Run("MachineSet", func(t *testing.T) {
-		testConfig := createMachineSetTestConfig(testNamespace, 1, 1, map[string]string{
+		testConfig := createMachineSetTestConfig(testNamespace, 1, map[string]string{
 			nodeGroupMinSizeAnnotationKey: "1",
 			nodeGroupMaxSizeAnnotationKey: "1",
 		})
@@ -634,7 +634,7 @@ func TestControllerNodeGroupForNodeWithPositiveScalingBounds(t *testing.T) {
 	})
 
 	t.Run("MachineDeployment", func(t *testing.T) {
-		testConfig := createMachineDeploymentTestConfig(testNamespace, 1, 1, map[string]string{
+		testConfig := createMachineDeploymentTestConfig(testNamespace, 1, map[string]string{
 			nodeGroupMinSizeAnnotationKey: "1",
 			nodeGroupMaxSizeAnnotationKey: "1",
 		})
@@ -666,14 +666,14 @@ func TestControllerNodeGroups(t *testing.T) {
 	assertNodegroupLen(t, controller, 0)
 
 	// Test #2: add 5 machineset-based nodegroups
-	machineSetConfigs := createMachineSetTestConfigs("MachineSet", 5, 1, 1, annotations)
+	machineSetConfigs := createMachineSetTestConfigs("MachineSet", 5, 1, annotations)
 	if err := addTestConfigs(t, controller, machineSetConfigs...); err != nil {
 		t.Fatalf("unexpected error: %v", err)
 	}
 	assertNodegroupLen(t, controller, 5)
 
 	// Test #2: add 2 machinedeployment-based nodegroups
-	machineDeploymentConfigs := createMachineDeploymentTestConfigs("MachineDeployment", 2, 1, 1, annotations)
+	machineDeploymentConfigs := createMachineDeploymentTestConfigs("MachineDeployment", 2, 1, annotations)
 	if err := addTestConfigs(t, controller, machineDeploymentConfigs...); err != nil {
 		t.Fatalf("unexpected error: %v", err)
 	}
@@ -697,14 +697,14 @@ func TestControllerNodeGroups(t *testing.T) {
 	}
 
 	// Test #5: machineset with no scaling bounds results in no nodegroups
-	machineSetConfigs = createMachineSetTestConfigs("MachineSet", 5, 1, 1, annotations)
+	machineSetConfigs = createMachineSetTestConfigs("MachineSet", 5, 1, annotations)
 	if err := addTestConfigs(t, controller, machineSetConfigs...); err != nil {
 		t.Fatalf("unexpected error: %v", err)
 	}
 	assertNodegroupLen(t, controller, 0)
 
 	// Test #6: machinedeployment with no scaling bounds results in no nodegroups
-	machineDeploymentConfigs = createMachineDeploymentTestConfigs("MachineDeployment", 2, 1, 1, annotations)
+	machineDeploymentConfigs = createMachineDeploymentTestConfigs("MachineDeployment", 2, 1, annotations)
 	if err := addTestConfigs(t, controller, machineDeploymentConfigs...); err != nil {
 		t.Fatalf("unexpected error: %v", err)
 	}
@@ -716,7 +716,7 @@ func TestControllerNodeGroups(t *testing.T) {
 	}
 
 	// Test #7: machineset with bad scaling bounds results in an error and no nodegroups
-	machineSetConfigs = createMachineSetTestConfigs("MachineSet", 5, 1, 1, annotations)
+	machineSetConfigs = createMachineSetTestConfigs("MachineSet", 5, 1, annotations)
 	if err := addTestConfigs(t, controller, machineSetConfigs...); err != nil {
 		t.Fatalf("unexpected error: %v", err)
 	}
@@ -725,7 +725,7 @@ func TestControllerNodeGroups(t *testing.T) {
 	}
 
 	// Test #8: machinedeployment with bad scaling bounds results in an error and no nodegroups
-	machineDeploymentConfigs = createMachineDeploymentTestConfigs("MachineDeployment", 2, 1, 1, annotations)
+	machineDeploymentConfigs = createMachineDeploymentTestConfigs("MachineDeployment", 2, 1, annotations)
 	if err := addTestConfigs(t, controller, machineDeploymentConfigs...); err != nil {
 		t.Fatalf("unexpected error: %v", err)
 	}
@@ -736,19 +736,27 @@ func TestControllerNodeGroups(t *testing.T) {
 
 func TestControllerNodeGroupsNodeCount(t *testing.T) {
 	type testCase struct {
-		nodeGroups    int
-		nodesPerGroup int
+		nodeGroups            int
+		nodesPerGroup         int
+		expectedNodeGroups    int
+		expectedNodesPerGroup int
 	}
 
 	var testCases = []testCase{{
-		nodeGroups:    0,
-		nodesPerGroup: 0,
+		nodeGroups:            0,
+		nodesPerGroup:         0,
+		expectedNodeGroups:    0,
+		expectedNodesPerGroup: 0,
 	}, {
-		nodeGroups:    1,
-		nodesPerGroup: 0,
+		nodeGroups:            1,
+		nodesPerGroup:         0,
+		expectedNodeGroups:    0,
+		expectedNodesPerGroup: 0,
 	}, {
-		nodeGroups:    2,
-		nodesPerGroup: 10,
+		nodeGroups:            2,
+		nodesPerGroup:         10,
+		expectedNodeGroups:    2,
+		expectedNodesPerGroup: 10,
 	}}
 
 	test := func(t *testing.T, tc testCase, testConfigs []*testConfig) {
@@ -759,8 +767,8 @@ func TestControllerNodeGroupsNodeCount(t *testing.T) {
 		if err != nil {
 			t.Fatalf("unexpected error: %v", err)
 		}
-		if got := len(nodegroups); got != tc.nodeGroups {
-			t.Fatalf("expected %d, got %d", tc.nodeGroups, got)
+		if got := len(nodegroups); got != tc.expectedNodeGroups {
+			t.Fatalf("expected %d, got %d", tc.expectedNodeGroups, got)
 		}
 
 		for i := range nodegroups {
@@ -768,8 +776,8 @@ func TestControllerNodeGroupsNodeCount(t *testing.T) {
 			if err != nil {
 				t.Fatalf("unexpected error: %v", err)
 			}
-			if got := len(nodes); got != tc.nodesPerGroup {
-				t.Fatalf("expected %d, got %d", tc.nodesPerGroup, got)
+			if got := len(nodes); got != tc.expectedNodesPerGroup {
+				t.Fatalf("expected %d, got %d", tc.expectedNodesPerGroup, got)
 			}
 		}
 	}
@@ -781,13 +789,201 @@ func TestControllerNodeGroupsNodeCount(t *testing.T) {
 
 	t.Run("MachineSet", func(t *testing.T) {
 		for _, tc := range testCases {
-			test(t, tc, createMachineSetTestConfigs(testNamespace, tc.nodeGroups, tc.nodesPerGroup, int32(tc.nodesPerGroup), annotations))
+			test(t, tc, createMachineSetTestConfigs(testNamespace, tc.nodeGroups, tc.nodesPerGroup, annotations))
 		}
 	})
 
 	t.Run("MachineDeployment", func(t *testing.T) {
 		for _, tc := range testCases {
-			test(t, tc, createMachineDeploymentTestConfigs(testNamespace, tc.nodeGroups, tc.nodesPerGroup, int32(tc.nodesPerGroup), annotations))
+			test(t, tc, createMachineDeploymentTestConfigs(testNamespace, tc.nodeGroups, tc.nodesPerGroup, annotations))
 		}
 	})
 }
+
+func TestControllerFindMachineFromNodeAnnotation(t *testing.T) {
+	testConfig := createMachineSetTestConfig(testNamespace, 1, map[string]string{
+		nodeGroupMinSizeAnnotationKey: "1",
+		nodeGroupMaxSizeAnnotationKey: "10",
+	})
+
+	controller, stop := mustCreateTestController(t, testConfig)
+	defer stop()
+
+	// Remove all the provider ID values on all the machines. We
+	// want to force findMachineByProviderID() to fallback to
+	// searching using the annotation on the node object.
+	for _, machine := range testConfig.machines {
+		machine.Spec.ProviderID = nil
+		if err := controller.machineInformer.Informer().GetStore().Update(machine); err != nil {
+			t.Fatalf("unexpected error updating machine, got %v", err)
+		}
+	}
+
+	// Test #1: Verify machine can be found from node annotation
+	machine, err := controller.findMachineByProviderID(testConfig.nodes[0].Spec.ProviderID)
+	if err != nil {
+		t.Fatalf("unexpected error: %v", err)
+	}
+	if machine == nil {
+		t.Fatal("expected to find machine")
+	}
+	if !reflect.DeepEqual(machine, testConfig.machines[0]) {
+		t.Fatalf("expected machines to be equal - expected %+v, got %+v", testConfig.machines[0], machine)
+	}
+
+	// Test #2: Verify machine is not found if it has no
+	// corresponding machine annotation.
+	node := testConfig.nodes[0].DeepCopy()
+	delete(node.Annotations, machineAnnotationKey)
+	if err := controller.nodeInformer.GetStore().Update(node); err != nil {
+		t.Fatalf("unexpected error updating node, got %v", err)
+	}
+	machine, err = controller.findMachineByProviderID(testConfig.nodes[0].Spec.ProviderID)
+	if err != nil {
+		t.Fatalf("unexpected error: %v", err)
+	}
+	if machine != nil {
+		t.Fatal("expected find to fail")
+	}
+}
+
+func TestControllerMachineSetNodeNamesWithoutLinkage(t *testing.T) {
+	testConfig := createMachineSetTestConfig(testNamespace, 3, map[string]string{
+		nodeGroupMinSizeAnnotationKey: "1",
+		nodeGroupMaxSizeAnnotationKey: "10",
+	})
+
+	controller, stop := mustCreateTestController(t, testConfig)
+	defer stop()
+
+	// Remove all linkage between node and machine.
+	for _, machine := range testConfig.machines {
+		machine.Spec.ProviderID = nil
+		if err := controller.machineInformer.Informer().GetStore().Update(machine); err != nil {
+			t.Fatalf("unexpected error updating machine, got %v", err)
+		}
+	}
+	for _, machine := range testConfig.machines {
+		machine.Status.NodeRef = nil
+		if err := controller.machineInformer.Informer().GetStore().Update(machine); err != nil {
+			t.Fatalf("unexpected error updating machine, got %v", err)
+		}
+	}
+
+	nodegroups, err := controller.nodeGroups()
+	if err != nil {
+		t.Fatalf("unexpected error: %v", err)
+	}
+
+	if l := len(nodegroups); l != 1 {
+		t.Fatalf("expected 1 nodegroup, got %d", l)
+	}
+
+	ng := nodegroups[0]
+	nodeNames, err := ng.Nodes()
+	if err != nil {
+		t.Fatalf("unexpected error: %v", err)
+	}
+
+	// We removed all linkage - so we should get 0 nodes back.
+	if len(nodeNames) != 0 {
+		t.Fatalf("expected len=0, got len=%v", len(nodeNames))
+	}
+}
+
+func TestControllerMachineSetNodeNamesUsingProviderID(t *testing.T) {
+	testConfig := createMachineSetTestConfig(testNamespace, 3, map[string]string{
+		nodeGroupMinSizeAnnotationKey: "1",
+		nodeGroupMaxSizeAnnotationKey: "10",
+	})
+
+	controller, stop := mustCreateTestController(t, testConfig)
+	defer stop()
+
+	// Remove Status.NodeRef.Name on all the machines. We want to
+	// force machineSetNodeNames() to only consider the provider
+	// ID for lookups.
+	for _, machine := range testConfig.machines {
+		machine.Status.NodeRef = nil
+		if err := controller.machineInformer.Informer().GetStore().Update(machine); err != nil {
+			t.Fatalf("unexpected error updating machine, got %v", err)
+		}
+	}
+
+	nodegroups, err := controller.nodeGroups()
+	if err != nil {
+		t.Fatalf("unexpected error: %v", err)
+	}
+
+	if l := len(nodegroups); l != 1 {
+		t.Fatalf("expected 1 nodegroup, got %d", l)
+	}
+
+	ng := nodegroups[0]
+	nodeNames, err := ng.Nodes()
+	if err != nil {
+		t.Fatalf("unexpected error: %v", err)
+	}
+
+	if len(nodeNames) != len(testConfig.nodes) {
+		t.Fatalf("expected len=%v, got len=%v", len(testConfig.nodes), len(nodeNames))
+	}
+
+	sort.Slice(nodeNames, func(i, j int) bool {
+		return nodeNames[i].Id < nodeNames[j].Id
+	})
+
+	for i := range testConfig.nodes {
+		if nodeNames[i].Id != testConfig.nodes[i].Spec.ProviderID {
+			t.Fatalf("expected %q, got %q", testConfig.nodes[i].Spec.ProviderID, nodeNames[i].Id)
+		}
+	}
+}
+
+func TestControllerMachineSetNodeNamesUsingStatusNodeRefName(t *testing.T) {
+	testConfig := createMachineSetTestConfig(testNamespace, 3, map[string]string{
+		nodeGroupMinSizeAnnotationKey: "1",
+		nodeGroupMaxSizeAnnotationKey: "10",
+	})
+
+	controller, stop := mustCreateTestController(t, testConfig)
+	defer stop()
+
+	// Remove all the provider ID values on all the machines. We
+	// want to force machineSetNodeNames() to fallback to
+	// searching using Status.NodeRef.Name.
+	for _, machine := range testConfig.machines {
+		machine.Spec.ProviderID = nil
+		if err := controller.machineInformer.Informer().GetStore().Update(machine); err != nil {
+			t.Fatalf("unexpected error updating machine, got %v", err)
+		}
+	}
+
+	nodegroups, err := controller.nodeGroups()
+	if err != nil {
+		t.Fatalf("unexpected error: %v", err)
+	}
+
+	if l := len(nodegroups); l != 1 {
+		t.Fatalf("expected 1 nodegroup, got %d", l)
+	}
+
+	nodeNames, err := nodegroups[0].Nodes()
+	if err != nil {
+		t.Fatalf("unexpected error: %v", err)
+	}
+
+	if len(nodeNames) != len(testConfig.nodes) {
+		t.Fatalf("expected len=%v, got len=%v", len(testConfig.nodes), len(nodeNames))
+	}
+
+	sort.Slice(nodeNames, func(i, j int) bool {
+		return nodeNames[i].Id < nodeNames[j].Id
+	})
+
+	for i := range testConfig.nodes {
+		if nodeNames[i].Id != testConfig.nodes[i].Spec.ProviderID {
+			t.Fatalf("expected %q, got %q", testConfig.nodes[i].Spec.ProviderID, nodeNames[i].Id)
+		}
+	}
+}
